<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Bootstrap;

// Import framework stuff
use Org\Mxchange\CoreFramework\Configuration\FrameworkConfiguration;
use Org\Mxchange\CoreFramework\Connection\Database\DatabaseConnection;
use Org\Mxchange\CoreFramework\Connector\Database\DatabaseConnector;
use Org\Mxchange\CoreFramework\Console\Tools\ConsoleTools;
use Org\Mxchange\CoreFramework\EntryPoint\ApplicationEntryPoint;
use Org\Mxchange\CoreFramework\Factory\Object\ObjectFactory;
use Org\Mxchange\CoreFramework\Generic\FrameworkInterface;
use Org\Mxchange\CoreFramework\Generic\NullPointerException;
use Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper;
use Org\Mxchange\CoreFramework\Localization\ManageableLanguage;
use Org\Mxchange\CoreFramework\Loader\ClassLoader;
use Org\Mxchange\CoreFramework\Manager\ManageableApplication;
use Org\Mxchange\CoreFramework\Middleware\Debug\DebugMiddleware;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;
use Org\Mxchange\CoreFramework\Request\Requestable;
use Org\Mxchange\CoreFramework\Response\Responseable;
use Org\Mxchange\CoreFramework\Utils\Strings\StringUtils;

// Import SPL stuff
use \BadMethodCallException;
use \InvalidArgumentException;
use \SplFileInfo;

/**
 * A framework-bootstrap class which helps the frameworks to bootstrap ... ;-)
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
final class FrameworkBootstrap {

	/**
	 * Detected server address
	 */
	private static $serverAddress = NULL;

	/**
	 * Instance of a Requestable class
	 */
	private static $requestInstance = NULL;

	/**
	 * Instance of a Responseable class
	 */
	private static $responseInstance = NULL;

	/**
	 * Instance of a FrameworkConfiguration class
	 */
	private static $configurationInstance = NULL;

	/**
	 * Database instance
	 */
	private static $databaseInstance = NULL;

	/**
	 * Language system instance
	 */
	private static $languageInstance = NULL;

	/*
	 * Includes applications may have. They will be tried in the given order,
	 * some will become soon deprecated.
	 */
	private static $configAppIncludes = [
		// The ApplicationHelper class (required)
		'class_ApplicationHelper' => 'required',
		// Some debugging stuff (optional but can be committed)
		'debug'                   => 'optional',
		// Application's exception handler (optional but can be committed)
		'exceptions'              => 'optional',
		// Application's configuration file (committed, non-local specific)
		'config'                  => 'required',
		// Local configuration file (optional, not committed, listed in .gitignore)
		'config-local'            => 'optional',
		// Application data (deprecated)
		'data'                    => 'deprecated',
		// Application loader (deprecated)
		'loader'                  => 'deprecated',
		// Application initializer (deprecated)
		'init'                    => 'deprecated',
		// Application starter (deprecated)
		'starter'                 => 'deprecated',
	];

	/**
	 * Detected application's name
	 */
	private static $detectedApplicationName;

	/**
	 * Detected application's full path
	 */
	private static $detectedApplicationPath;

	/**
	 * Private constructor, no instance is needed from this class as only
	 * static methods exist.
	 */
	private function __construct () {
		// Prevent making instances from this "utilities" class
	}

	/**
	 * Some "getter" for a configuration instance, making sure, it is unique
	 *
	 * @return	$configurationInstance	An instance of a FrameworkConfiguration class
	 */
	public static function getConfigurationInstance () {
		// Is the instance there?
		if (is_null(self::$configurationInstance)) {
			// Init new instance
			self::$configurationInstance = new FrameworkConfiguration();
		}

		// Return it
		return self::$configurationInstance;
	}

	/**
	 * Getter for detected application name
	 *
	 * @return	$detectedApplicationName	Detected name of application
	 */
	public static function getDetectedApplicationName () {
		return self::$detectedApplicationName;
	}

	/**
	 * Getter for detected application's full path
	 *
	 * @return	$detectedApplicationPath	Detected full path of application
	 */
	public static function getDetectedApplicationPath () {
		return self::$detectedApplicationPath;
	}

	/**
	 * "Getter" to get response/request type from analysis of the system.
	 *
	 * @return	$requestType	Analyzed request type
	 */
	public static function getRequestTypeFromSystem () {
		// Default is console
		$requestType = 'console';

		// Is 'HTTP_HOST' set?
		if (isset($_SERVER['HTTP_HOST'])) {
			// Then it is a HTML response/request.
			$requestType = 'html';
		}

		// Return it
		return $requestType;
	}

	/**
	 * Checks whether the given file/path is in open_basedir(). This does not
	 * gurantee that the file is actually readable and/or writeable. If you need
	 * such gurantee then please use isReadableFile() instead.
	 *
	 * @param	$fileInstance	An instance of a SplFileInfo class
	 * @return	$isReachable	Whether it is within open_basedir()
	 */
	public static function isReachableFilePath (SplFileInfo $fileInstance) {
		// Is not reachable by default
		//* NOISY-DEBUG: */ printf('[%s:%d]: fileInstance=%s - CALLED!' . PHP_EOL, __METHOD__, __LINE__, get_class($fileInstance));
		$isReachable = false;

		// Get open_basedir parameter
		$openBaseDir = trim(ini_get('open_basedir'));

		// Is it set?
		//* NOISY-DEBUG: */ printf('[%s:%d]: openBaseDir=%s' . PHP_EOL, __METHOD__, __LINE__, $openBaseDir);
		if (!empty($openBaseDir)) {
			// Check all entries
			foreach (explode(PATH_SEPARATOR, $openBaseDir) as $dir) {
				// Check on existence
				//* NOISY-DEBUG: */ printf('[%s:%d]: dir=%s' . PHP_EOL, __METHOD__, __LINE__, $dir);
				if (substr($fileInstance->getPathname(), 0, strlen($dir)) == $dir) {
					// Is reachable
					$isReachable = true;

					// Abort lookup as it has been found in open_basedir
					//* NOISY-DEBUG: */ printf('[%s:%d]: BREAK!' . PHP_EOL, __METHOD__, __LINE__);
					break;
				}
			}
		} else {
			// If open_basedir is not set, all is allowed
			//* NOISY-DEBUG: */ printf('[%s:%d]: All is allowed - BREAK!' . PHP_EOL, __METHOD__, __LINE__);
			$isReachable = true;
		}

		// Return status
		//* NOISY-DEBUG: */ printf('[%s:%d]: isReachable=%d - EXIT' . PHP_EOL, __METHOD__, __LINE__, intval($isReachable));
		return $isReachable;
	}

	/**
	 * Checks whether the give file is within open_basedir() (done by
	 * isReachableFilePath()), is actually a file and is readable.
	 *
	 * @param	$fileInstance	An instance of a SplFileInfo class
	 * @return	$isReadable		Whether the file is readable (and therefor exists)
	 */
	public static function isReadableFile (SplFileInfo $fileInstance) {
		// Check if it is a file and readable
		//* NOISY-DEBUG: */ printf('[%s:%d]: fileInstance=%s - CALLED!' . PHP_EOL, __METHOD__, __LINE__, get_class($fileInstance));
		$isReadable = (
			(
				self::isReachableFilePath($fileInstance)
			) && (
				$fileInstance->isFile()
			) && (
				$fileInstance->isReadable()
			)
		);

		// Return status
		//* NOISY-DEBUG: */ printf('[%s:%d]: isReadable=%d - EXIT!' . PHP_EOL, __METHOD__, __LINE__, intval($isReadable));
		return $isReadable;
	}

	/**
	 * Loads given include file
	 *
	 * @param	$fileInstance	An instance of a SplFileInfo class
	 * @return	void
	 * @throws	InvalidArgumentException	If file was not found or not readable or deprecated
	 */
	public static function loadInclude (SplFileInfo $fileInstance) {
		// Should be there ...
		//* NOISY-DEBUG: */ printf('[%s:%d]: fileInstance=%s - CALLED!' . PHP_EOL, __METHOD__, __LINE__, $fileInstance->__toString());
		if (!self::isReadableFile($fileInstance)) {
			// Abort here
			throw new InvalidArgumentException(sprintf('Cannot find fileInstance.pathname=%s.', $fileInstance->getPathname()));
		} elseif (!$fileInstance->isFile()) {
			// Not a file
			throw new InvalidArgumentException(sprintf('fileInstance.pathname=%s is not a file', $fileInstance->getPathname()));
		} elseif (substr($fileInstance->__toString(), -4, 4) != '.php') {
			// Not PHP script
			throw new InvalidArgumentException(sprintf('fileInstance.pathname=%s is not a PHP script', $fileInstance->getPathname()));
		}

		// Load it
		require_once $fileInstance->getPathname();

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Does the actual bootstrap. I think the amount of statically loaded
	 * include files cannot be reduced here as those files are need to early
	 * in the bootstrap phase. If you can find an other solution than this, with
	 * lesser "static includes" (means not loaded by the class loader), please
	 * let me know.
	 *
	 * @return	void
	 */
	public static function doBootstrap () {
		// Load basic include files to continue bootstrapping
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		self::loadInclude(new SplFileInfo(sprintf('%smain%sinterfaces%sclass_FrameworkInterface.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR)));
		self::loadInclude(new SplFileInfo(sprintf('%smain%sclasses%sclass_BaseFrameworkSystem.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR)));
		self::loadInclude(new SplFileInfo(sprintf('%smain%sclasses%sutils%sstrings%sclass_StringUtils.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR)));
		self::loadInclude(new SplFileInfo(sprintf('%smain%sinterfaces%sregistry%sclass_Registerable.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR)));
		self::loadInclude(new SplFileInfo(sprintf('%sconfig%sclass_FrameworkConfiguration.php', ApplicationEntryPoint::detectFrameworkPath(), DIRECTORY_SEPARATOR)));

		// Load global configuration
		self::loadInclude(new SplFileInfo(sprintf('%s%s', ApplicationEntryPoint::detectFrameworkPath(), 'config-global.php')));

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Initializes the framework by scanning for all framework-relevant
	 * classes, interfaces and exception. Then determine the request type and
	 * initialize a Requestable instance which will then contain all request
	 * parameter, also from CLI. Next step is to validate the application
	 * (very basic).
	 *
	 * @return	void
	 */
	public static function initFramework () {
		/**
		 * 1) Load class loader and scan framework classes, interfaces and
		 *    exceptions.
		 */
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		self::scanFrameworkClasses();

		/*
		 * 2) Determine the request type, console or web and store request and
		 *    response here. This also initializes the request instance with
		 *    all given parameters (see doc-tag for possible sources of
		 *    parameters).
		 */
		self::determineRequestType();

		/*
		 * 3) Now, that there are all request parameters being available, check
		 *    if 'application' is supplied. If it is not found, abort execution, if
		 *    found, continue below with next step.
		 */
		self::validateApplicationParameter();

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Initializes the detected application. This may fail if required files
	 * are not found in the application's base path (not to be confused with
	 * 'application_base_path' which only points to /some/foo/application/.
	 *
	 * @return	void
	 */
	public static function prepareApplication () {
		/*
		 * Now check and load all files, found deprecated files will throw a
		 * warning at the user.
		 */
		//* NOISY-DEBUG: */ printf('[%s:%d]: self::configAppIncludes()=%d - CALLED!' . PHP_EOL, __METHOD__, __LINE__, count(self::$configAppIncludes));
		foreach (self::$configAppIncludes as $fileName => $status) {
			// Construct file instance
			//* NOISY-DEBUG: */ printf('[%s:%d]: fileName=%s,status=%s' . PHP_EOL, __METHOD__, __LINE__, $fileName, $status);
			$fileInstance = new SplFileInfo(sprintf('%s%s.php', self::getDetectedApplicationPath(), $fileName));

			// Determine if this file is wanted/readable/deprecated
			if (($status == 'required') && (!self::isReadableFile($fileInstance))) {
				// Nope, required file cannot be found/read from
				ApplicationEntryPoint::exitApplication(sprintf('Application "%s" does not have required file "%s.php". Please add it.', self::getDetectedApplicationName(), $fileInstance->getBasename()));
			} elseif (($fileInstance->isFile()) && (!$fileInstance->isReadable())) {
				// Found, not readable file
				ApplicationEntryPoint::exitApplication(sprintf('File "%s.php" from application "%s" cannot be read. Please fix CHMOD.', $fileInstance->getBasename(), self::getDetectedApplicationName()));
			} elseif (($status != 'required') && (!self::isReadableFile($fileInstance))) {
				// Not found but optional/deprecated file, skip it
				//* NOISY-DEBUG: */ printf('[%s:%d]: fileName=%s,status=%s - SKIPPED!' . PHP_EOL, __METHOD__, __LINE__, $fileName, $status);
				continue;
			}

			// Is the file deprecated?
			if ($status == 'deprecated') {
				// Issue warning
				trigger_error(sprintf('Deprecated file "%s.php" found, will not load it to avoid problems. Please remove it from your application "%s" to avoid this warning.', $fileName, self::getDetectedApplicationName()), E_USER_WARNING);

				// Skip loading deprecated file
				continue;
			}

			// Load it
			//* NOISY-DEBUG: */ printf('[%s:%d]: Invoking self::loadInclude(%s) ...' . PHP_EOL, __METHOD__, __LINE__, $fileInstance->__toString());
			self::loadInclude($fileInstance);
		}

		// After this, sort the configuration array
		self::getConfigurationInstance()->sortConfigurationArray();

		// Scan for application's classes, exceptions and interfaces
		ClassLoader::scanApplicationClasses();

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Starts a fully initialized application, the class ApplicationHelper must
	 * be loaded at this point.
	 *
	 * @return	void
	 */
	public static function startApplication () {
		// Is there an application helper instance?
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		$applicationInstance = call_user_func_array(
			[
				'Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper', 'getSelfInstance'
			], []
		);

		// Some sanity checks
		//* NOISY-DEBUG: */ printf('[%s:%d]: applicationInstance=%s' . PHP_EOL, __METHOD__, __LINE__, $applicationInstance->__toString());
		if ((empty($applicationInstance)) || (is_null($applicationInstance))) {
			// Something went wrong!
			ApplicationEntryPoint::exitApplication(sprintf('[Main:] The application <span class="app_name">%s</span> could not be launched because the helper class <span class="class_name">%s</span> is not loaded.',
				self::getDetectedApplicationName(),
				'Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper'
			));
		} elseif (!is_object($applicationInstance)) {
			// No object!
			ApplicationEntryPoint::exitApplication(sprintf('[Main:] The application <span class="app_name">%s</span> could not be launched because &#39;app&#39; is not an object (%s).',
				self::getDetectedApplicationName(),
				gettype($applicationInstance)
			));
		} elseif (!($applicationInstance instanceof ManageableApplication)) {
			// Missing interface
			ApplicationEntryPoint::exitApplication(sprintf('[Main:] The application <span class="app_name">%s</span> could not be launched because &#39;app&#39; is lacking required interface ManageableApplication.',
				self::getDetectedApplicationName()
			));
		}

		// Now call all methods in one go
		//* NOISY-DEBUG: */ printf('[%s:%d]: Initializing application ...' . PHP_EOL, __METHOD__, __LINE__);
		foreach (['setupApplicationData', 'initApplication', 'launchApplication'] as $methodName) {
			// Call method
			//*NOISY-DEBUG: */ printf('[%s:%d]: Invoking methodName=%s ...' . PHP_EOL, __METHOD__, __LINE__, $methodName);
			call_user_func([$applicationInstance, $methodName]);
		}

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Initializes database instance, no need to double-call this method
	 *
	 * @return	void
	 * @throws	BadMethodCallException	If this method was invoked twice
	 */
	public static function initDatabaseInstance () {
		// Get application instance
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		$applicationInstance = ApplicationHelper::getSelfInstance();

		// Is the database instance already set?
		//* NOISY-DEBUG: */ printf('[%s:%d]: self::databaseInstance[]=%s' . PHP_EOL, __METHOD__, __LINE__, gettype(self::getDatabaseInstance()));
		if (self::getDatabaseInstance() instanceof DatabaseConnector) {
			// Yes, then abort here
			throw new BadMethodCallException('Method called twice.');
		}

		// Initialize database layer
		$databaseInstance = ObjectFactory::createObjectByConfiguredName(self::getConfigurationInstance()->getConfigEntry('database_type') . '_class');

		// Prepare database instance
		$connectionInstance = DatabaseConnection::createDatabaseConnection($databaseInstance);

		// Set it in application helper
		//* NOISY-DEBUG: */ printf('[%s:%d]: Setting connectionInstance=%s ...' . PHP_EOL, __METHOD__, __LINE__, $connectionInstance->__toString());
		self::setDatabaseInstance($connectionInstance);

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Detects the server address (SERVER_ADDR) and set it in configuration
	 *
	 * @return	$serverAddress	The detected server address
	 * @throws	UnknownHostnameException	If SERVER_NAME cannot be resolved to an IP address
	 * @todo	Have to check some more entries from $_SERVER here
	 */
	public static function detectServerAddress () {
		// Is the entry set?
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		if (!isset(self::$serverAddress)) {
			// Is it set in $_SERVER?
			if (!empty($_SERVER['SERVER_ADDR'])) {
				// Set it from $_SERVER
				//* NOISY-DEBUG: */ printf('[%s:%d]: Setting self::serverAddress=%s from SERVER_ADDR ...' . PHP_EOL, __METHOD__, __LINE__, $_SERVER['SERVER_ADDR']);
				self::$serverAddress = $_SERVER['SERVER_ADDR'];
			} elseif (isset($_SERVER['SERVER_NAME'])) {
				// Resolve IP address
				//* NOISY-DEBUG: */ printf('[%s:%d]: Resolving SERVER_NAME=%s ...' . PHP_EOL, __METHOD__, __LINE__, $_SERVER['SERVER_NAME']);
				$serverIp = ConsoleTools::resolveIpAddress($_SERVER['SERVER_NAME']);

				// Is it valid?
				//* NOISY-DEBUG: */ printf('[%s:%d]: serverId[%s]=%s' . PHP_EOL, __METHOD__, __LINE__, gettype($serverIp), $serverIp);
				if ($serverIp === false) {
					/*
					 * Why is gethostbyname() returning the host name and not
					 * false as many other PHP functions are doing? ;-(
					 */
					throw new UnknownHostnameException(sprintf('Cannot resolve "%s" to an IP address. Please fix your setup.', $_SERVER['SERVER_NAME']));
				}

				// Al fine, set it
				//* NOISY-DEBUG: */ printf('[%s:%d]: Setting self::serverAddress=%s from resolver ...' . PHP_EOL, __METHOD__, __LINE__, $serverAddress);
				self::$serverAddress = $serverIp;
			} else {
				// Run auto-detecting through console tools lib
				//* NOISY-DEBUG: */ printf('[%s:%d]: Invoking ConsoleTools::acquireSelfIpAddress() ...' . PHP_EOL, __METHOD__, __LINE__);
				self::$serverAddress = ConsoleTools::acquireSelfIpAddress();
			}
		}

		// Return it
		//* NOISY-DEBUG: */ printf('[%s:%d]: self::serverAddress=%s - EXIT!' . PHP_EOL, __METHOD__, __LINE__, self::$serverAddress);
		return self::$serverAddress;
	}

	/**
	 * Setter for default time zone (must be correct!)
	 *
	 * @param	$timezone	The timezone string (e.g. Europe/Berlin)
	 * @return	$success	If timezone was accepted
	 * @throws	InvalidArgumentException	If $timezone is empty
	 */
	public static function setDefaultTimezone (string $timezone) {
		// Is it set?
		//* NOISY-DEBUG: */ printf('[%s:%d]: timezone=%s - CALLED!' . PHP_EOL, __METHOD__, __LINE__, $timezone);
		if (empty($timezone)) {
			// Entry is empty
			throw new InvalidArgumentException('Parameter "timezone" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// Default success
		$success = FALSE;

		/*
		 * Set desired time zone to prevent date() and related functions to
		 * issue an E_WARNING.
		 */
		$success = date_default_timezone_set($timezone);

		// Return status
		//* NOISY-DEBUG: */ printf('[%s:%d]: success=%d - EXIT!' . PHP_EOL, __METHOD__, __LINE__, intval($success));
		return $success;
	}

	/**
	 * Checks whether HTTPS is set in $_SERVER
	 *
	 * @return	$isset	Whether HTTPS is set
	 * @todo	Test more fields
	 */
	public static function isHttpSecured () {
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		return (
			(
				(
					isset($_SERVER['HTTPS'])
				) && (
					strtolower($_SERVER['HTTPS']) == 'on'
				)
			) || (
				(
					isset($_SERVER['HTTP_X_FORWARDED_PROTO'])
				) && (
					strtolower($_SERVER['HTTP_X_FORWARDED_PROTO']) == 'https'
				)
			)
		);
	}

	/**
	 * Dectect and return the base URL for all URLs and forms
	 *
	 * @return	$baseUrl	Detected base URL
	 */
	public static function detectBaseUrl () {
		// Initialize the URL
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		$protocol = 'http';

		// Do we have HTTPS?
		if (self::isHttpSecured()) {
			// Add the >s< for HTTPS
			$protocol = 'https';
		}

		// Construct the full URL and secure it against CSRF attacks
		$baseUrl = sprintf('%s://%s%s', $protocol, self::detectDomain(), self::detectScriptPath());

		// Return the URL
		//* NOISY-DEBUG: */ printf('[%s:%d]: baseUrl=%s - EXIT!' . PHP_EOL, __METHOD__, __LINE__, $baseUrl);
		return $baseUrl;
	}

	/**
	 * Detect safely and return the full domain where this script is installed
	 *
	 * @return	$fullDomain		The detected full domain
	 */
	public static function detectDomain () {
		// Full domain is localnet.invalid by default
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		$fullDomain = 'localnet.invalid';

		// Is the server name there?
		if (isset($_SERVER['SERVER_NAME'])) {
			// Detect the full domain
			$fullDomain = htmlentities(strip_tags($_SERVER['SERVER_NAME']), ENT_QUOTES);
		}

		// Return it
		//* NOISY-DEBUG: */ printf('[%s:%d]: fullDomain=%s - EXIT!' . PHP_EOL, __METHOD__, __LINE__, $fullDomain);
		return $fullDomain;
	}

	/**
	 * Detect safely the script path without trailing slash which is the glue
	 * between "http://your-domain.invalid/" and "script-name.php"
	 *
	 * @return	$scriptPath		The script path extracted from $_SERVER['SCRIPT_NAME']
	 */
	public static function detectScriptPath () {
		// Default is empty
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		$scriptPath = '';

		// Is the scriptname set?
		if (isset($_SERVER['SCRIPT_NAME'])) {
			// Get dirname from it and replace back-slashes with slashes for lame OSes...
			$scriptPath = str_replace("\\", '/', dirname($_SERVER['SCRIPT_NAME']));
		}

		// Return it
		//* NOISY-DEBUG: */ printf('[%s:%d]: scriptPath=%s - EXIT!' . PHP_EOL, __METHOD__, __LINE__, $scriptPath);
		return $scriptPath;
	}

	/**
	 * 1) Loads class scanner and scans all framework's classes and interfaces.
	 * This method also registers the class loader's method autoLoad() for the
	 * SPL auto-load feature. Remember that you can register additional methods
	 * (not functions, please) for other libraries.
	 *
	 * Yes, I know about Composer, but I like to keep my class loader around.
	 * You can always use mine as long as your classes have a namespace
	 * according naming-convention: Vendor\Project\Group[\SubGroup]
	 *
	 * @return	void
	 */
	private static function scanFrameworkClasses () {
		// Include class loader
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		require self::getConfigurationInstance()->getConfigEntry('framework_base_path') . 'loader/class_ClassLoader.php';

		// Register auto-load function with the SPL
		spl_autoload_register('Org\Mxchange\CoreFramework\Loader\ClassLoader::autoLoad');

		// Scan for all framework classes, exceptions and interfaces
		ClassLoader::scanFrameworkClasses();

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * 2) Determines request/response type and stores the created
	 * request/response instances in this object for later usage.
	 *
	 * @return	void
	 */
	private static function determineRequestType () {
		// Determine request type
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		$request = self::getRequestTypeFromSystem();
		$requestType = self::getRequestTypeFromSystem();

		// Create a new request object
		//* NOISY-DEBUG: */ printf('[%s:%d]: request=%s,requestType=%s' . PHP_EOL, __METHOD__, __LINE__, $request, $requestType);
		$requestInstance = ObjectFactory::createObjectByName(sprintf('Org\Mxchange\CoreFramework\Request\%sRequest', StringUtils::convertToClassName($request)));

		// Remember request instance here
		//* NOISY-DEBUG: */ printf('[%s:%d]: Setting requestInstance=%s ...' . PHP_EOL, __METHOD__, __LINE__, $requestInstance->__toString());
		self::setRequestInstance($requestInstance);

		// Do we have another response?
		if ($requestInstance->isRequestElementSet('request')) {
			// Then use it
			$request = strtolower($requestInstance->getRequestElement('request'));
			$requestType = $request;
		}

		// ... and a new response object
		//* NOISY-DEBUG: */ printf('[%s:%d]: request=%s,requestType=%s - AFTER!' . PHP_EOL, __METHOD__, __LINE__, $request, $requestType);
		$responseInstance = ObjectFactory::createObjectByName(sprintf('Org\Mxchange\CoreFramework\Response\%sResponse', StringUtils::convertToClassName($request)));

		// Remember response instance here
		//* NOISY-DEBUG: */ printf('[%s:%d]: Setting responseInstance=%s ...' . PHP_EOL, __METHOD__, __LINE__, $responseInstance->__toString());
		self::setResponseInstance($responseInstance);

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * 3) Validate parameter 'application' if it is set and the application is there.
	 *
	 * @return	void
	 */
	private static function validateApplicationParameter () {
		// Is the parameter set?
		//* NOISY-DEBUG: */ printf('[%s:%d]: CALLED!' . PHP_EOL, __METHOD__, __LINE__);
		if (!self::getRequestInstance()->isRequestElementSet('app')) {
			/*
			 * Don't continue here, the application 'selector' is no longer
			 * supported and only existed as an idea to select the proper
			 * application (by user).
			 */
			ApplicationEntryPoint::exitApplication('No application specified. Please provide a parameter "app" and retry.');
		}

		// Get it for local usage
		$applicationName = self::getRequestInstance()->getRequestElement('app');

		// Secure it, by keeping out tags
		//* NOISY-DEBUG: */ printf('[%s:%d]: applicationName=%s' . PHP_EOL, __METHOD__, __LINE__, $applicationName);
		$applicationName = htmlentities(strip_tags($applicationName), ENT_QUOTES);

		// Secure it a little more with a reg.exp.
		//* NOISY-DEBUG: */ printf('[%s:%d]: applicationName=%s' . PHP_EOL, __METHOD__, __LINE__, $applicationName);
		$applicationName = preg_replace('/([^a-z0-9_-])+/i', '', $applicationName);

		// Construct FQPN (Full-Qualified Path Name) for ApplicationHelper class
		//* NOISY-DEBUG: */ printf('[%s:%d]: applicationName=%s' . PHP_EOL, __METHOD__, __LINE__, $applicationName);
		$applicationPath = sprintf(
			'%s%s%s',
			self::getConfigurationInstance()->getConfigEntry('application_base_path'),
			$applicationName,
			DIRECTORY_SEPARATOR
		);

		// Full path for application
		// Is the path there? This secures a bit the parameter (from untrusted source).
		//* NOISY-DEBUG: */ printf('[%s:%d]: applicationPath=%s' . PHP_EOL, __METHOD__, __LINE__, $applicationPath);
		if ((!is_dir($applicationPath)) || (!is_readable($applicationPath))) {
			// Not found or not readable
			ApplicationEntryPoint::exitApplication(sprintf('Application "%s" not found.', $applicationName));
		}

		// Set the detected application's name and full path for later usage
		//* NOISY-DEBUG: */ printf('[%s:%d]: Setting applicationPath=%s,applicationName=%s' . PHP_EOL, __METHOD__, __LINE__, $applicationPath, $applicationName);
		self::$detectedApplicationPath = $applicationPath;
		self::$detectedApplicationName = $applicationName;

		// Trace message
		//* NOISY-DEBUG: */ printf('[%s:%d]: EXIT!' . PHP_EOL, __METHOD__, __LINE__);
	}

	/**
	 * Getter for request instance
	 *
	 * @return	$requestInstance	An instance of a Requestable class
	 */
	public static function getRequestInstance () {
		return self::$requestInstance;
	}

	/**
	 * Getter for response instance
	 *
	 * @return	$responseInstance	An instance of a Responseable class
	 */
	public static function getResponseInstance () {
		return self::$responseInstance;
	}

	/**
	 * Setter for request instance
	 *
	 * @param	$requestInstance	An instance of a Requestable class
	 * @return	void
	 */
	private static function setRequestInstance (Requestable $requestInstance) {
		self::$requestInstance = $requestInstance;
	}

	/**
	 * Setter for response instance
	 *
	 * @param	$responseInstance	An instance of a Responseable class
	 * @return	void
	 */
	private static function setResponseInstance (Responseable $responseInstance) {
		self::$responseInstance = $responseInstance;
	}

	/**
	 * Setter for database instance
	 *
	 * @param	$databaseInstance	An instance of a DatabaseConnection class
	 * @return	void
	 */
	public static function setDatabaseInstance (DatabaseConnection $databaseInstance) {
		self::$databaseInstance = $databaseInstance;
	}

	/**
	 * Getter for database instance
	 *
	 * @return	$databaseInstance	An instance of a DatabaseConnection class
	 */
	public static function getDatabaseInstance () {
		// Return instance
		return self::$databaseInstance;
	}

	/**
	 * Private getter for language instance
	 *
	 * @return	$languageInstance	An instance of a ManageableLanguage class
	 */
	public static function getLanguageInstance () {
		return self::$languageInstance;
	}

	/**
	 * Setter for language instance
	 *
	 * @param	$languageInstance	An instance of a ManageableLanguage class
	 * @return	void
	 */
	public static function setLanguageInstance (ManageableLanguage $languageInstance) {
		self::$languageInstance = $languageInstance;
	}

}
