<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Command;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Factory\Object\ObjectFactory;
use Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper;
use Org\Mxchange\CoreFramework\Generic\FrameworkInterface;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;
use Org\Mxchange\CoreFramework\Request\Requestable;
use Org\Mxchange\CoreFramework\Response\Responseable;
use Org\Mxchange\CoreFramework\Traits\Resolver\ResolverTrait;
use Org\Mxchange\CoreFramework\Traits\Template\CompileableTemplateTrait;

// Import SPL stuff
use \InvalidArgumentException;

/**
 * A general (base) command
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
abstract class BaseCommand extends BaseFrameworkSystem {
	// Load traits
	use CompileableTemplateTrait;
	use ResolverTrait;

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct (string $className) {
		// Call parent constructor
		parent::__construct($className);
	}

	/**
	 * Initializes the template engine
	 *
	 * @param	$templateType	Type of template, e.g. 'html', 'image', 'console' ...
	 * @return	void
	 */
	public final function initTemplateEngine (string $templateType) {
		// Check paramter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('BASE-COMMAND: templateType=%s - CALLED!', $templateType));
		if (empty($templateType)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "templateType" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// Prepare a template instance
		$templateInstance = ObjectFactory::createObjectByConfiguredName(sprintf('%s_template_class', $templateType));

		// Set it here
		$this->setTemplateInstance($templateInstance);

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-COMMAND: EXIT!');
	}

	/**
	 * Sends a generic HTTP response with header, menu, content and footer
	 *
	 * @param	$requestInstance	An instance of a class with an Requestable interface
	 * @param	$responseInstance	An instance of a class with an Responseable interface
	 * @param	$suffix				Optional template suffix, e.g. '_form' for forms
	 * @return	void
	 */
	protected function sendGenericGetResponse (Requestable $requestInstance, Responseable $responseInstance, string $suffix = '') {
		// This command doesn't handle any POST requests, so only handle get request
		assert(!$requestInstance->isPostRequestMethod());

		// Get the application instance
		$applicationInstance = ApplicationHelper::getSelfInstance();

		// Transfer application data
		$this->getTemplateInstance()->assignApplicationData();

		// Assign base URL
		$this->getTemplateInstance()->assignConfigVariable('base_url');

		// Load the master template
		$masterTemplate = $applicationInstance->buildMasterTemplateName();

		// Load header template
		$this->getTemplateInstance()->loadCodeTemplate('header');

		// Compile and assign it with a variable
		$this->getTemplateInstance()->compileTemplate();
		$this->getTemplateInstance()->assignTemplateWithVariable('header', 'header');

		// Load footer template
		$this->getTemplateInstance()->loadCodeTemplate('footer');

		// Compile and assign it with a variable
		$this->getTemplateInstance()->compileTemplate();
		$this->getTemplateInstance()->assignTemplateWithVariable('footer', 'footer');

		// Load the content template
		$this->getTemplateInstance()->loadCodeTemplate($this->getResolverInstance()->getCommandName() . $suffix);

		// Assign the content template with the master template as a content ... ;)
		$this->getTemplateInstance()->assignTemplateWithVariable($applicationInstance->getAppShortName() . '_' . $this->getResolverInstance()->getCommandName(), 'main_content');

		// Load the master template
		$this->getTemplateInstance()->loadCodeTemplate($masterTemplate);

		// Set title
		$this->getTemplateInstance()->assignVariable('title', FrameworkBootstrap::getLanguageInstance()->getMessage('page_' . $applicationInstance->getAppShortName() . '_' . $this->getResolverInstance()->getCommandName() . '_title'));

		// Construct the menu in every command. We could do this in BaseCommand class. But this means
		// *every* command has a navigation system and that is want we don't want.
		$menuInstance = ObjectFactory::createObjectByConfiguredName($applicationInstance->getAppShortName() . '_' . $this->getResolverInstance()->getCommandName() . '_menu_class');

		// Render the menu
		$menuInstance->renderMenu();

		// Transfer it to the template engine instance
		$menuInstance->transferContentToTemplateEngine($this->getTemplateInstance());

		/*
		 * ... and all variables. This should be merged together in a pattern
		 * to make things easier. A cache mechanism should be added between
		 * these two calls to cache compiled templates.
		 */
		$this->getTemplateInstance()->compileVariables();

		// Get the content back from the template engine and put it in response class
		$this->getTemplateInstance()->transferToResponse($responseInstance);
	}

}
