<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Result\Database;

// Import framework stuff
use Org\Mxchange\CoreFramework\Criteria\Local\LocalSearchCriteria;
use Org\Mxchange\CoreFramework\Criteria\Local\LocalUpdateCriteria;
use Org\Mxchange\CoreFramework\Criteria\Storing\StoreableCriteria;
use Org\Mxchange\CoreFramework\Database\Frontend\DatabaseFrontend;
use Org\Mxchange\CoreFramework\Database\Backend\BaseDatabaseBackend;
use Org\Mxchange\CoreFramework\Result\Database\BaseDatabaseResult;
use Org\Mxchange\CoreFramework\Result\Search\SearchableResult;
use Org\Mxchange\CoreFramework\Result\Update\UpdateableResult;

// Import SPL stuff
use \InvalidArgumentException;
use \OutOfBoundsException;
use \SeekableIterator;

/**
 * A database result class
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class CachedDatabaseResult extends BaseDatabaseResult implements SearchableResult, UpdateableResult, SeekableIterator {
	// Exception constants
	const EXCEPTION_INVALID_DATABASE_RESULT = 0x1c0;
	const EXCEPTION_RESULT_UPDATE_FAILED    = 0x1c1;

	/**
	 * Current position in array
	 */
	private $currentPos = -1;

	/**
	 * Current row
	 */
	private $currentRow = NULL;

	/**
	 * Result array
	 */
	private $resultArray = [];

	/**
	 * Array of out-dated entries
	 */
	private $outDated = [];

	/**
	 * Affected rows
	 */
	private $affectedRows = 0;

	/**
	 * Found value
	 */
	private $foundValue = '';

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: CONSTRUCTED!');
		parent::__construct(__CLASS__);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: EXIT!');
	}

	/**
	 * Creates an instance of this result by a provided result array
	 *
	 * @param	$resultArray		The array holding the result from query
	 * @return	$resultInstance		An instance of this class
	 * @throws	InvalidArgumentException	If a parameter is invalid
	 */
	public static final function createCachedDatabaseResult (array $resultArray) {
		// Misses an element?
		//* DEBUG-DIE: */ die(sprintf('[%s:%d]: resultArray=%s', __METHOD__, __LINE__, print_r($resultArray, true)));
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: resultArray()=%d - CALLED!', count($resultArray)));
		if (count($resultArray) == 0) {
			// Cannot be empty
			throw new InvalidArgumentException('Array "resultArray" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		} elseif (!array_key_exists(BaseDatabaseResult::RESULT_NAME_ROWS, $resultArray)) {
			// Yes, then abort here
			throw new InvalidArgumentException(sprintf('resultArray(%d)=%s has no element "%s".', count($resultArray), print_r($resultArray, TRUE), BaseDatabaseResult::RESULT_NAME_ROWS));
		} elseif (!array_key_exists(BaseDatabaseResult::RESULT_NAME_STATUS, $resultArray)) {
			// Yes, then abort here
			throw new InvalidArgumentException(sprintf('resultArray(%d)=%s has no element "%s".', count($resultArray), print_r($resultArray, TRUE), BaseDatabaseResult::RESULT_NAME_STATUS));
		}

		// Get a new instance
		$resultInstance = new CachedDatabaseResult();

		// Set the result array
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: Setting resultArray()=%d ...', count($resultArray)));
		$resultInstance->setResultArray($resultArray);

		// Reset current position
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: Invoking resultInstance->resetCurrentPosition() ...');
		$resultInstance->resetCurrentPosition();

		// Set affected rows
		$resultInstance->setAffectedRows(count($resultArray[BaseDatabaseResult::RESULT_NAME_ROWS]));

		// Return the instance
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: resultInstance=%s - EXIT!', $resultInstance->__toString()));
		return $resultInstance;
	}

	/**
	 * Setter for result array
	 *
	 * @param	$resultArray	The array holding the result from query
	 * @return	void
	 */
	protected final function setResultArray (array $resultArray) {
		$this->resultArray = $resultArray;
	}

	/**
	 * Updates the current entry by given update criteria
	 *
	 * @param	$updateInstance		An instance of an Updateable criteria
	 * @return	void
	 */
	private function updateCurrentEntryByCriteria (LocalUpdateCriteria $updateInstance) {
		// Get the current entry key
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: updateInstance=%s - CALLED!', $updateInstance->__toString()));
		$entryKey = $this->key();

		// Now get the update criteria array and update all entries
		foreach ($updateInstance->getUpdateCriteria() as $criteriaKey => $criteriaValue) {
			// Update data
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: criteriaKey=%s,criteriaValue[%s]=%s', $criteriaKey, gettype($criteriaValue), $criteriaValue));
			$this->resultArray[BaseDatabaseResult::RESULT_NAME_ROWS][$entryKey][$criteriaKey] = $criteriaValue;

			// Mark it as out-dated
			$this->outDated[$criteriaKey] = 1;
		}

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: EXIT!');
	}

	/**
	 * "Iterator" method next() to advance to the next valid entry. This method
	 * does also check if result is invalid
	 *
	 * @return	$nextValid	Whether the next entry is valid
	 */
	public function next () {
		// Default is not valid
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: CALLED!');
		$nextValid = false;

		// Increase position
		$this->currentPos++;

		// Is the result valid?
		if ($this->valid()) {
			// Next entry found, so cache it
			$this->currentRow = $this->resultArray[BaseDatabaseResult::RESULT_NAME_ROWS][$this->currentPos];
			$nextValid = true;
		}

		// Return the result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: nextValid=%d - EXIT!', intval($nextValid)));
		return $nextValid;
	}

	/**
	 * Seeks for to a specified position
	 *
	 * @param	$seekPosition	Position to seek to
	 * @return	void
	 * @throws	OutOfBoundsException	If the position is not seekable
	 */
	public function seek (int $seekPosition) {
		// Validate parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: seekPosition=%d - CALLED!', $seekPosition));
		if ($seekPosition < 0) {
			// Throw exception
			throw new OutOfBoundsException(sprintf('seekPositon=%d is not seekable', $seekPosition));
		}

		// Rewind to beginning
		$this->rewind();

		// Search for the entry
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: this->currentPos=%d,seekPosition=%d', $this->currentPos, $seekPosition));
		while (($this->currentPos < $seekPosition) && ($this->valid())) {
			// Continue on
			$this->next();
		}

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: EXIT!');
	}

	/**
	 * Gives back the current position or null if not found
	 *
	 * @return	$current	Current element to give back
	 */
	public function current () {
		// Default is not found
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: CALLED!');
		$current = NULL;

		// Does the current enty exist?
		if (isset($this->resultArray[BaseDatabaseResult::RESULT_NAME_ROWS][$this->currentPos])) {
			// Then get it
			$current = $this->resultArray[BaseDatabaseResult::RESULT_NAME_ROWS][$this->currentPos];
		}

		// Return the result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: current[]=%s - EXIT!', gettype($current)));
		return $current;
	}

	/**
	 * Checks if next() and rewind will give a valid result
	 *
	 * @return	$isValid Whether the next/rewind entry is valid
	 */
	public function valid () {
		// Check if all is fine ...
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: this->currentPos=%d - CALLED!', $this->currentPos));
		$isValid = ($this->ifStatusIsOkay() && isset($this->resultArray[BaseDatabaseResult::RESULT_NAME_ROWS][$this->currentPos]) && isset($this->resultArray[BaseDatabaseResult::RESULT_NAME_ROWS][0]));

		// Return the result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: isValid=%d - EXIT!', intval($isValid)));
		return $isValid;
	}

	/**
	 * Returns count of entries
	 *
	 * @return	$isValid Whether the next/rewind entry is valid
	 */
	public function count () {
		// Count rows
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: CALLED!');
		$count = count($this->resultArray[BaseDatabaseResult::RESULT_NAME_ROWS]);

		// Return it
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: count=%d - EXIT!', $count));
		return $count;
	}

	/**
	 * Determines whether the status of the query was fine (BaseDatabaseBackend::RESULT_OKAY)
	 *
	 * @return	$ifStatusOkay	Whether the status of the query was okay
	 */
	public function ifStatusIsOkay () {
		// Check all conditions
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: this->currentPos=%d - CALLED!', $this->currentPos));
		$ifStatusOkay = (isset($this->resultArray[BaseDatabaseResult::RESULT_NAME_STATUS]) && $this->resultArray[BaseDatabaseResult::RESULT_NAME_STATUS] === BaseDatabaseBackend::RESULT_OKAY);

		// Return status
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: ifStatusOkay=%s - EXIT!', intval($ifStatusOkay)));
		return $ifStatusOkay;
	}

	/**
	 * Gets the current key of iteration
	 *
	 * @return	$currentPos	Key from iterator
	 */
	public function key () {
		// Return current array position
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: this->currentPos=%d - CALLED!', $this->currentPos));
		return $this->currentPos;
	}

	/**
	 * Rewind to the beginning and clear array $currentRow
	 *
	 * @return	void
	 */
	public function rewind () {
		// Reset both current array position and current row
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: this->currentPos=%d - CALLED!', $this->currentPos));
		$this->resetCurrentPosition();
		$this->currentRow = [];

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: EXIT!');
	}

	/**
	 * Resets current array position to 0 if at least one record is there or -1
	 * if no record is there.
	 *
	 * @return	void
	 */
	private function resetCurrentPosition () {
		// Reset position
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: CALLED!');
		$this->currentPos = ($this->count() > 0 ? 0 : -1);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: this->currentPos=%d - EXIT!', $this->currentPos));
	}

	/**
	 * Searches for an entry in data result and returns it
	 *
	 * @param	$criteriaInstance	The criteria to look inside the data set
	 * @return	$result				Found result entry
	 * @todo	0% done
	 */
	public function searchEntry (LocalSearchCriteria $criteriaInstance) {
		// Unfinished code
		$this->debugBackTrace(sprintf('[%s:%d]: criteriaInstance=%s', __METHOD__, __LINE__, print_r($criteriaInstance, TRUE)));
	}

	/**
	 * Adds an update request to the database result for writing it to the
	 * database layer
	 *
	 * @param	$updateInstance	An instance of a updateable criteria
	 * @return	void
	 * @throws	ResultUpdateException	If no result was updated
	 */
	public function add2UpdateQueue (LocalUpdateCriteria $updateInstance) {
		// Rewind the pointer
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: updateInstance=%s - CALLED!', $updateInstance->__toString()));
		$this->rewind();

		// Get search criteria
		$searchInstance = $updateInstance->getSearchInstance();

		// Get affected rows
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: searchInstance=%s', $searchInstance->__toString()));
		$foundEntries = $this->getAffectedRows();

		// And start looking for the result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: foundEntries=%d,searchInstance->limit=%d', $foundEntries, $searchInstance->getLimit()));
		while (($this->valid()) && ($foundEntries < $searchInstance->getLimit())) {
			// Get next entry
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: Invoking this->next() ...');
			$this->next();

			// Get current entry
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: Invoking this->current() ...');
			$currentEntry = $this->current();

			// Is this entry found?
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: currentEntry[]=%s - EXIT!', gettype($currentEntry)));
			if ($searchInstance->ifEntryMatches($currentEntry)) {
				// Update this entry
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: Invoking this->updateCurrentEntryByCriteria(%s) ...', $updateInstance->__toString()));
				$this->updateCurrentEntryByCriteria($updateInstance);

				// Count one up
				$foundEntries++;
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: foundEntries=%d', $foundEntries));
			}
		}

		// If no entry is found/updated throw an exception
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: foundEntries=%d', $foundEntries));
		if ($foundEntries == 0) {
			// Throw an exception here
			throw new ResultUpdateException($this, self::EXCEPTION_RESULT_UPDATE_FAILED);
		}

		// Set affected rows
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: Invoking this->setAffectedRows(%d) ...', $foundEntries));
		$this->setAffectedRows($foundEntries);

		// Set update instance
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: Invoking this->setUpdateInstance(%s) ...', $updateInstance->__toString()));
		$this->setUpdateInstance($updateInstance);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: EXIT!');
	}

	/**
	 * Setter for affected rows
	 *
	 * @param	$rows	Number of affected rows
	 * @return	void
	 */
	public final function setAffectedRows (int $rows) {
		$this->affectedRows = $rows;
	}

	/**
	 * Getter for affected rows
	 *
	 * @return	$rows	Number of affected rows
	 */
	public final function getAffectedRows () {
		return $this->affectedRows;
	}

	/**
	 * Getter for found value of previous found() call
	 *
	 * @return	$foundValue		Found value of previous found() call
	 */
	public final function getFoundValue () {
		return $this->foundValue;
	}

	/**
	 * Checks whether we have out-dated entries or not
	 *
	 * @return	$needsUpdate	Whether we have out-dated entries
	 */
	public function ifDataNeedsFlush () {
		// Check if records are out-dated
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: CALLED!');
		$needsUpdate = (count($this->outDated) > 0);

		// Return it
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: needsUpdate=%d - EXIT!', intval($needsUpdate)));
		return $needsUpdate;
	}

	/**
	 * Adds registration elements to a given dataset instance
	 *
	 * @param	$criteriaInstance	An instance of a StoreableCriteria class
	 * @return	void
	 */
	public function addElementsToDataSet (StoreableCriteria $criteriaInstance) {
		// Walk only through out-dated columns
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: criteriaInstance=%s - CALLED!', $criteriaInstance->__toString()));
		foreach ($this->outDated as $key => $dummy) {
			// Does this key exist?
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: key=%s,dummy[]=%s', $key, gettype($dummy)));
			if ($this->find($key)) {
				// Then update it
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: Invoking criteriaInstance->addCriteria(%s,foundValue[]=%s) ...', $key, gettype($this->getFoundValue())));
				$criteriaInstance->addCriteria($key, $this->getFoundValue());
			}
		}

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: EXIT!');
	}

	/**
	 * Find a key inside the result array
	 *
	 * @param	$key	The key we shall find
	 * @return	$found	Whether the key was found or not
	 * @throws	InvalidArgumentException	If a parameter is invalid
	 */
	public function find (string $key) {
		// Check parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: key=%s - CALLED!', $key));
		if (empty($key)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "key" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// By default nothing is found
		$found = false;

		// Rewind the pointer
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: Invoking this->rewind() ...');
		$this->rewind();

		// Walk through all entries
		while ($this->valid()) {
			// Advance to next entry
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: Invoking this->next() ...');
			$this->next();

			// Get the whole array
			$currentEntry = $this->current();

			// Is the element there?
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: currentEntry[]=%s - EXIT!', gettype($currentEntry)));
			if (isset($currentEntry[$key])) {
				// Okay, found!
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: key=%s is found in currentEntry', $key));
				$found = true;

				// So "cache" it
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: Setting this->foundValue to currentEntry[%s][]=%s is found in currentEntry', $key, gettype($currentEntry[$key])));
				$this->foundValue = $currentEntry[$key];

				// And stop searching
				//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage('CACHED-DATABASE-RESULT: BREAK!');
				break;
			}
		}

		// Return the result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: found=%d - EXIT!', intval($found)));
		return $found;
	}

	/**
	 * Solver for result index value with call-back method
	 *
	 * @param	$databaseColumn		Database column where the index might be found
	 * @param	$frontendInstance	The frontend instance to ask for array element
	 * @para	$callback			Call-back object for setting the index;
	 * 								0=object instance,1=method name
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is invalid
	 * @todo	Find a caching way without modifying the result array
	 */
	public function solveResultIndex (string $databaseColumn, DatabaseFrontend $frontendInstance, array $callback) {
		// Check parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: databaseColumn=%s,frontendInstance=%s,callback()=%d - CALLED!', $databaseColumn, $frontendInstance->__toString(), count($callback)));
		if (empty($key)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "key" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		} elseif (count($callback) != 2) {
			// Throw it again
			throw new InvalidArgumentException(sprintf('callback()=%d must be exactly 2', count($callback)), FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		} elseif (!is_callable($callback)) {
			// Throw it again
			throw new InvalidArgumentException(sprintf('%s:%s() is not callable', $callback[0], $callback[1]), FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// By default nothing is found
		$indexValue = 0;

		// Is the element in result itself found?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: Invoking this->find(%s),frontendInstance->indexKey=%s ...', $databaseColumn, $frontendInstance->getIndexKey()));
		if ($this->find($databaseColumn)) {
			// Use this value
			$indexValue = $this->getFoundValue();
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: indexValue[]=%s from this->find(%s) - databaseColumn', gettype($indexValue), $databaseColumn));
		} elseif ($this->find($frontendInstance->getIndexKey())) {
			// Use this value
			$indexValue = $this->getFoundValue();
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('CACHED-DATABASE-RESULT: indexValue[]=%s from this->find(%s) - frontendInstance->indexKey', gettype($indexValue), $frontendInstance->getIndexKey()));
		}

		// Set the index
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('CACHED-DATABASE-RESULT: Invoking %s:%s(indexValue[]=%s) ...', $callback[0], $callback[1], gettype($indexValue)));
		call_user_func_array($callback, [$indexValue]);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('CACHED-DATABASE-RESULT: EXIT!');
	}

}
