<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Filter\Validator\Username;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Filter\BaseFilter;
use Org\Mxchange\CoreFramework\Filter\Chain\FilterChainException;
use Org\Mxchange\CoreFramework\Filter\Filterable;
use Org\Mxchange\CoreFramework\Registry\Object\ObjectRegistry;
use Org\Mxchange\CoreFramework\Request\Requestable;
use Org\Mxchange\CoreFramework\Response\Responseable;
use Org\Mxchange\CoreFramework\User\UsernameMissingException;

/**
 * A concrete filter for validating the username. This filter may intercept the
 * filter chain if no username is given or if supplied username has an invalid
 * form. It could also intercept our filter chain if username is already taken.
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class UserNameValidatorFilter extends BaseFilter implements Filterable {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this filter class
	 *
	 * @return	$filterInstance			An instance of this filter class
	 */
	public static final function createUserNameValidatorFilter () {
		// Get a new instance
		$filterInstance = new UserNameValidatorFilter();

		// Return the instance
		return $filterInstance;
	}

	/**
	 * Executes the filter with given request and response objects
	 *
	 * @param	$requestInstance	An instance of a class with an Requestable interface
	 * @param	$responseInstance	An instance of a class with an Responseable interface
	 * @return	void
	 * @throws	FilterChainException	If this filter fails to operate
	 */
	public function execute (Requestable $requestInstance, Responseable $responseInstance) {
		// Get username from request
		$userName = $requestInstance->getRequestElement('username');

		// Is the username set?
		if (is_null($userName)) {
			// Not found in form so stop the filtering process
			$requestInstance->setIsRequestValid(FALSE);

			// Add a message to the response
			$responseInstance->addFatalMessage('username_unset');

			// Abort here
			throw new FilterChainException($this, self::EXCEPTION_FILTER_CHAIN_INTERCEPTED);
		} elseif (empty($userName)) {
			// Empty field!
			$requestInstance->setIsRequestValid(FALSE);

			// Add a message to the response
			$responseInstance->addFatalMessage('username_empty');

			// Abort here
			throw new FilterChainException($this, self::EXCEPTION_FILTER_CHAIN_INTERCEPTED);
		} elseif ($this->ifUserNameIsTaken($userName)) {
			// Username is already taken
			$requestInstance->setIsRequestValid(FALSE);

			// Add a message to the response
			$responseInstance->addFatalMessage('username_taken');

			// Abort here
			throw new FilterChainException($this, self::EXCEPTION_FILTER_CHAIN_INTERCEPTED);
		}
	}

	/**
	 * Check whether the username as already been taken
	 *
	 * @param	$userName		Username to check for existence
	 * @return	$alreadyTaken	Whether the username has been taken
	 */
	private function ifUserNameIsTaken (string $userName) {
		// Default is already taken
		$alreadyTaken = true;

		// Initialize instance
		$userInstance = NULL;

		// Get a registry instance
		$registryInstance = ObjectRegistry::getRegistry('generic');

		// Is the user already there?
		if ($registryInstance->instanceExists('user')) {
			// Use the instance for checking for the email
			$userInstance = $registryInstance->getInstance('user');
			$userInstance->setUserName($userName);
		} else {
			// If this instance is created then the username *does* exist
			try {
				// Get a new instance
				$userInstance = call_user_func_array(array(FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('user_class'), 'createMemberByUsername'), array($userName));

				// Remember this user instance in our registry for later usage
				$registryInstance->addInstance('user', $userInstance);
			} catch (UsernameMissingException $e) {
				// User was not found
			}
		}

		// Does the username exist?
		if ((is_null($userInstance)) || ($userInstance->ifUsernameExists() === false)) {
			// This username is still available
			$alreadyTaken = false;
		}

		// Return the result
		return $alreadyTaken;
	}

}
