<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Parser\Xml;

// Import framework stuff
use Org\Mxchange\CoreFramework\Generic\FrameworkInterface;
use Org\Mxchange\CoreFramework\Middleware\Debug\DebugMiddleware;
use Org\Mxchange\CoreFramework\Parser\BaseParser;
use Org\Mxchange\CoreFramework\Parser\Parseable;
use Org\Mxchange\CoreFramework\Template\CompileableTemplate;

// Import SPL stuff
use \InvalidArgumentException;

/**
 * A Xml Parser class
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class XmlParser extends BaseParser implements Parseable {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of the class Parser and prepares it for usage
	 *
	 * @param	$templateInstance	A CompileableTemplate instance
	 * @return	$parserInstance		An instance of this parser
	 */
	public static final function createXmlParser (CompileableTemplate $templateInstance) {
		// Get a new instance
		$parserInstance = new XmlParser();

		// Set the template instance
		$parserInstance->setTemplateInstance($templateInstance);

		// Return the prepared instance
		return $parserInstance;
	}

	/**
	 * Parses the given XML content
	 *
	 * @param	$content	Valid XML content
	 * @return	void
	 * @throws	XmlParserException	If an XML error was found
	 */
	public function parseXmlContent (string $content) {
		// Convert all to UTF8
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('XML-PARSER: content()=%d - CALLED!', strlen($content)));
		if (empty($content)) {
			// No empty content
			throw new InvalidArgumentException('content is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		} elseif (function_exists('recode')) {
			// Recode found, so use it
			$content = recode('html..utf8', $content);
		} elseif (function_exists('mb_convert_encoding')) {
			// Use mb_convert_encoding()
			$content = mb_convert_encoding($content, 'UTF-8', 'auto');
		} else {
			// @TODO We need to find a fallback solution here
			DebugMiddleware::getSelfInstance()->partialStub('Cannot find recode/mbstring extension!');
		}

		// Get an XML parser
		$xmlParser = xml_parser_create();

		// Force case-folding to on
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('XML-PARSER: Created xmlParser=%s ...', $xmlParser));
		xml_parser_set_option($xmlParser, XML_OPTION_CASE_FOLDING, TRUE);

		// Set UTF-8
		xml_parser_set_option($xmlParser, XML_OPTION_TARGET_ENCODING, 'UTF-8');

		// Set object
		xml_set_object($xmlParser, $this->getTemplateInstance());

		// Set handler call-backs
		xml_set_element_handler($xmlParser, 'startElement', 'finishElement');
		xml_set_character_data_handler($xmlParser, 'characterHandler');

		// Now parse the XML tree
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('XML-PARSER: Invoking xml_parse(%s, content()=%d) ...', $xmlParser, strlen($content)));
		if (!xml_parse($xmlParser, $content)) {
			// Error found in XML
			//* DEBUG: */ exit(__METHOD__ . ':<pre>'.htmlentities($content).'</pre>');
			throw new XmlParserException([$this, $xmlParser], Parseable::EXCEPTION_XML_PARSER_ERROR);
		}

		// Free the parser
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('XML-PARSER: Freeing xmlParser=%s ...', $xmlParser));
		xml_parser_free($xmlParser);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('XML-PARSER: EXIT!');
	}

}
