<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Registry;

// Import framework stuff
use Org\Mxchange\CoreFramework\Factory\Object\ObjectFactory;
use Org\Mxchange\CoreFramework\Generic\FrameworkInterface;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;
use Org\Mxchange\CoreFramework\Traits\Iterator\IteratorTrait;

// Import SPL stuff
use \InvalidArgumentExeption;
use \IteratorAggregate;
use \UnexpectedValueException;

/**
 * A general Registry
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
abstract class BaseRegistry extends BaseFrameworkSystem implements Register, Registerable, IteratorAggregate {
	// Load traits
	use IteratorTrait;

	/**
	 * Glue for generating a registry key
	 */
	const REGISTRY_KEY_GLUE = '_';

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct (string $className) {
		// Call parent constructor
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: className=%s - CONSTRUCTED!', $className));
		parent::__construct($className);

		// Init generic arrays
		$this->initGenericArrayGroup('registry', 'generic');

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('BASE-REGISTRY: EXIT!');
	}

	/**
	 * Returns an iterator for this whole registry.
	 *
	 * @param	$onlyRegistries		Only iterate on these sub-registry keys, default is all
	 * @return	$iteratorInstance	An instance of a Iterator class
	 */
	public function getIterator (array $onlyRegistries = []) {
		// Get iterator
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: onlyRegistries()=%d - CALLED!', count($onlyRegistries)));
		$iteratorInstance = $this->getIteratorInstance();

		// Is it set?
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('BASE-REGISTRY: iteratorInstance[]=%s', gettype($iteratorInstance)));
		if (is_null($iteratorInstance)) {
			// Then instance it
			$iteratorInstance = ObjectFactory::createObjectByConfiguredName('registry_iterator_class', [$this]);

			// ... and set it here
			$this->setIteratorInstance($iteratorInstance);
		}

		// Init iterator instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: Invoking iteratorInstance->initIterator(onlyRegistries()=%d) ...', count($onlyRegistries)));
		$iteratorInstance->initIterator($onlyRegistries);

		// Return it
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: iteratorInstance=%s - EXIT!', $iteratorInstance->__toString()));
		return $iteratorInstance;
	}

	/**
	 * Getter for whole generic registry
	 *
	 * @return	$instanceRegistry	The whole generic registry array
	 */
	public final function getGenericRegistry () {
		return $this->getGenericSubArray('registry', 'generic');
	}

	/**
	 * Adds a new entry to the given list name. If you want to add objects
	 * please use addInstance() and getInstance() instead.
	 *
	 * @param	$key	The key to identify the whole list
	 * @param	$value	The value to be stored
	 * @return	void
	 * @throws	InvalidArgumentException	If a paramter has an invalid value
	 */
	public final function addEntry (string $key, $value) {
		// Check parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: key=%s,value[]=%s - CALLED!', $key, gettype($value)));
		if (empty($key)) {
			// Throw IAE
			throw new InvalidArgumentExeption('Parameter "key" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// Push it
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: Invoking this->pushValueToGenericArrayKey(registry,generic,%s,value[]=%s) ...', $key, gettype($value)));
		$this->pushValueToGenericArrayKey('registry', 'generic', $key, $value);

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('BASE-REGISTRY: EXIT!');
	}

	/**
	 * Getter for entries or "sub entries"
	 *
	 * @param	$key	Key
	 * @return	$entries	An array with entries from this registry
	 * @throws	InvalidArgumentException	If a paramter has an invalid value
	 */
	public final function getEntries (string $key = NULL) {
		// Check parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: key[%s]=%s - CALLED!', gettype($key), $key));
		if (!is_null($key) && empty($key)) {
			// Throw IAE
			throw new InvalidArgumentExeption('Parameter "key" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// Default is whole array
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('BASE-REGISTRY: Invoking this->getGenericArray(registry) ...');
		$entries = $this->getGenericArray('registry');

		// Is $key set?
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('BASE-REGISTRY: key[]=%s,entries()=%d', $key, count($entries)));
		if (!is_null($key)) {
			// Then use this entry
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: Invoking this->getGenericArrayKey(registry,generic,%s) ...', $key));
			$entries = $this->getGenericArrayKey('registry', 'generic', $key);
		}

		// Return the array
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: entries()=%d - EXIT!', count($entries)));
		return $entries;
	}

	/**
	 * "Getter" for an array of all entries for given key
	 *
	 * @param	$arrayKey	The array (key) to look in
	 * @param	$lookFor	The key to look for
	 * @return	$entry		An array with all keys
	 * @throws	InvalidArgumentException	If a paramter has an invalid value
	 * @throws	UnexpectedValueException	If $value3 is not an array
	 */
	public function getArrayFromKey (string $arrayKey, string $lookFor) {
		// Check parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: arrayKey=%s,lookFor=%s - CALLED!', $arrayKey, $lookFor));
		if (empty($arrayKey)) {
			// Throw IAE
			throw new InvalidArgumentExeption('Parameter "arrayKey" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		} elseif (empty($lookFor)) {
			// Throw IAE
			throw new InvalidArgumentExeption('Parameter "lookFor" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// Init array
		$entry = [];

		// "Walk" over all entries
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('BASE-REGISTRY: Checking arrayKey=%s,lookFor=%s', $arrayKey, $lookFor));
		foreach ($this->getEntries($arrayKey) as $key => $value) {
			// If $value matches the $lookFor, we need to look for more entries for this!
			/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('BASE-REGISTRY: Checking key=%s,value=%s,lookFor=%s', $key, $value, $lookFor));
			if ($lookFor == $value) {
				// Look for more entries
				foreach ($this->getEntries() as $key2 => $value2) {
					// Now "walk" through all entries, if an array is returned
					/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('BASE-REGISTRY: key2=%s,value2[]=%s', $key2, gettype($value2)));
					if (is_array($value2)) {
						// "Walk" through all of them
						/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('BASE-REGISTRY: Checking key2=%s,value2()=%s,lookFor=%s', $key2, count($value2), $lookFor));
						foreach ($value2 as $key3 => $value3) {
							// $value3 needs to be an array
							if (!is_array($value3)) {
								// Throw exception
								throw new UnexpectedValueException(sprintf('arrayKey=%s,key=%s,value=%s,key2=%s,key3=%s has unexpected value3[]=%s', $arrayKey, $key, $value, $key2, $key3, gettype($value3)), FrameworkInterface::EXCEPTION_UNEXPECTED_VALUE);
							}

							// Both keys must match!
							/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('BASE-REGISTRY: Checking key=%s,key3=%s,isset()=%s ...', $key, $key3, intval(isset($value3[$key]))));
							if (($key == $key3) || (isset($value3[$key]))) {
								// Then add it
								/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('BASE-REGISTRY: Adding value3[%s]=%s ...', $key, $value3[$key]));
								$entry[$key3] = $value3[$key];
							}
						}
					}
				}

				// Skip further lookups
				/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage('BASE-REGISTRY: BREAK!');
				break;
			}
		}

		// Return it
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: entry()=%d - EXIT!', count($entry)));
		return $entry;
	}

	/**
	 * "Getter" for a registry key for given prefix and array. This method
	 * calls implode() to get a suitable key. This method does not care about
	 * the indexes.
	 *
	 * @param	$prefix			Prefix for the key
	 * @param	$data			An array with data
	 * @return	$registryKey	A registry key
	 */
	public static function getRegistryKeyFromArray (string $prefix, array $data) {
		// Check parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: prefix=%s,data()=%d - CALLED!', $prefix, count($data)));
		if (empty($prefix)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "prefix" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		} elseif (count($data) == 0) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "data" is an empty array', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// "Generate" the key
		$registryKey = $prefix . self::REGISTRY_KEY_GLUE . implode(self::REGISTRY_KEY_GLUE, $data);

		// Return it
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('BASE-REGISTRY: registryKey=%s - EXIT!', $registryKey));
		return $registryKey;
	}

}
