<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Registry\Object;

// Import framework stuff
use Org\Mxchange\CoreFramework\Generic\NullPointerException;
use Org\Mxchange\CoreFramework\Registry\BaseRegistry;
use Org\Mxchange\CoreFramework\Registry\Registerable;

// Import SPL stuff
use \InvalidArgumentException;

/**
 * A registry for several data types and objects. Objects should be added by
 * addInstance() and therefore must implement the interface Registerable.
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class ObjectRegistry extends BaseRegistry implements ObjectRegister {
	/**
	 * Instances of this class
	 */
	private static $registryInstances = [];

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('OBJECT-REGISTRY: CONSTRUCTED!');
		parent::__construct(__CLASS__);

		// Init own array
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('OBJECT-REGISTRY: Invoking this->initGenericArrayGroup(registry,instance) ...');
		$this->initGenericArrayGroup('registry', 'instance');

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('OBJECT-REGISTRY: EXIT!');
	}

	/**
	 * Singleton getter for self instance. This class has no factory pattern
	 * because here is no need for special parameters.
	 *
	 * @param	$key	Key for for this instance
	 * @return	$registryInstance	Instance of this class
	 * @throws	InvalidArgumentException	If a parameter has an invalid value
	 */
	public static final function getRegistry (string $key) {
		// Check parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: key=%s - CALLED!', $key));
		if (empty($key)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "key" is empty');
		}

		// Is an instance there?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugMessage(sprintf('OBJECT-REGISTRY: self::registryInstance[%s]?=%d', $key, intval(isset(self::$registryInstances[$key]))));
		if (!isset(self::$registryInstances[$key])) {
			// Not yet, so create one
			self::$registryInstances[$key] = new ObjectRegistry();
		}

		// Return the instance
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: self::registryInstance[%s]=%s - EXIT!', $key, self::$registryInstances[$key]->__toString()));
		return self::$registryInstances[$key];
	}

	/**
	 * Checks whether an instance key was found
	 *
	 * @param	$instanceKey	The key holding an instance in registry
	 * @return	$exists			Whether the key exists in registry
	 * @throws	InvalidArgumentException	If a paramter has an invalid value
	 */
	public function instanceExists (string $instanceKey) {
		// Check parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: instanceKey=%s - CALLED!', $instanceKey));
		if (empty($instanceKey)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "instanceKey" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// Does this key exists?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: Invoking this->isGenericArrayKeySet(registry, instance, %s) ...', $instanceKey));
		$exists = $this->isGenericArrayKeySet('registry', 'instance', $instanceKey);

		// Return the result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: exists=%d - EXIT!', intval($exists)));
		return $exists;
	}

	/**
	 * Getter for whole instance registry
	 *
	 * @return	$instanceRegistry	The whole instance registry array
	 */
	public final function getInstanceRegistry () {
		return $this->getGenericSubArray('registry', 'instance');
	}

	/**
	 * Adds/overwrites a new instance to the registry at the given key
	 *
	 * @param	$instanceKey		The key to identify the instance
	 * @param	$objectInstance		An instance we shall store
	 * @return	void
	 * @throws	InvalidArgumentException	If a paramter has an invalid value
	 */
	public function addInstance (string $instanceKey, Registerable $objectInstance) {
		// Check parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: instanceKey=%s,objectInstance=%s - CALLED!', $instanceKey, $objectInstance->__toString()));
		if (empty($instanceKey)) {
			// Throw IAE
			throw new InvalidArgumentExeption('Parameter "instanceKey" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// Set entry in generic array
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: Invoking this->setGenericArrayKey(registry,instance,%s,%s) ...', $instanceKey, $objectInstance->__toString()));
		$this->setGenericArrayKey('registry', 'instance', $instanceKey, $objectInstance);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage('OBJECT-REGISTRY: EXIT!');
	}

	/**
	 * Gets a registered instance or null if not found
	 *
	 * @param	$instanceKey		The key to identify the instance
	 * @return	$objectInstance		An instance we shall store
	 * @throws	InvalidArgumentException	If a paramter has an invalid value
	 * @throws	NullPointerException	If the requested key is not found
	 */
	public function getInstance (string $instanceKey) {
		// Check parameter
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: instanceKey=%s - CALLED!', $instanceKey));
		if (empty($instanceKey)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "instanceKey" is empty', FrameworkInterface::EXCEPTION_INVALID_ARGUMENT);
		}

		// Is the instance there?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: Invoking this->instanceExists(%s) ...', $instanceKey));
		if (!$this->instanceExists($instanceKey)) {
			// This might happen if a non-registered key was requested
			throw new NullPointerException($this, self::EXCEPTION_IS_NULL_POINTER);
		}

		// Get the instance
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: Invoking this->getGenericArrayKey(registry,instance,%s) ...', $instanceKey));
		$objectInstance = $this->getGenericArrayKey('registry', 'instance', $instanceKey);

		// Return the result
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->traceMessage(sprintf('OBJECT-REGISTRY: objectInstance=%s - EXIT!', $objectInstance->__toString()));
		return $objectInstance;
	}

}
