<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Template\Engine;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Factory\Object\ObjectFactory;
use Org\Mxchange\CoreFramework\Filesystem\InvalidDirectoryException;
use Org\Mxchange\CoreFramework\Generic\FrameworkInterface;
use Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper;
use Org\Mxchange\CoreFramework\Image\BaseImage;
use Org\Mxchange\CoreFramework\Middleware\Debug\DebugMiddleware;
use Org\Mxchange\CoreFramework\Parser\Parseable;
use Org\Mxchange\CoreFramework\Response\Responseable;
use Org\Mxchange\CoreFramework\Template\CompileableTemplate;
use Org\Mxchange\CoreFramework\Template\Engine\BaseTemplateEngine;
use Org\Mxchange\CoreFramework\Utils\Strings\StringUtils;

// Import SPL stuff
use \SplFileInfo;
use \UnexpectedValueException;

/**
 * The own template engine for loading caching and sending out images
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2023 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class ImageTemplateEngine extends BaseTemplateEngine implements CompileableTemplate {
	/**
	 * Main nodes in the XML tree ('image' is ignored)
	 */
	private $mainNodes = array(
		'base',
		'type',
		'resolution',
		'background-color',
		'foreground-color',
		'image-string'
	);

	/**
	 * Sub nodes in the XML tree
	 */
	private $subNodes = array(
		'name',
		'string-name',
		'x',
		'y',
		'font-size',
		'width',
		'height',
		'red',
		'green',
		'blue',
		'text'
	);

	/**
	 * Current main node
	 */
	private $currMainNode = '';

	/**
	 * Instance of the image
	 */
	private $imageInstance = NULL;

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of the class TemplateEngine and prepares it for usage
	 *
	 * @return	$templateInstance		An instance of TemplateEngine
	 * @throws	UnexpectedValueException		If the provided $templateBasePath is empty or no string
	 * @throws	InvalidDirectoryException	If $templateBasePath is no
	 *											directory or not found
	 * @throws	BasePathReadProtectedException	If $templateBasePath is
	 *											read-protected
	 */
	public static final function createImageTemplateEngine () {
		// Get a new instance
		$templateInstance = new ImageTemplateEngine();

		// Get the application instance from registry
		$applicationInstance = ApplicationHelper::getSelfInstance();

		// Determine base path
		$templateBasePath = FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('application_base_path') . $applicationInstance->getAppShortName(). '/';

		// Is the base path valid?
		if (empty($templateBasePath)) {
			// Base path is empty
			throw new UnexpectedValueException(sprintf('[%s:%d] Variable templateBasePath is empty.', $templateInstance->__toString(), __LINE__), FrameworkInterface::EXCEPTION_UNEXPECTED_VALUE);
		} elseif (!is_dir($templateBasePath)) {
			// Is not a path
			throw new InvalidDirectoryException(array($templateInstance, $templateBasePath), self::EXCEPTION_INVALID_PATH_NAME);
		} elseif (!is_readable($templateBasePath)) {
			// Is not readable
			throw new BasePathReadProtectedException(array($templateInstance, $templateBasePath), self::EXCEPTION_READ_PROTECED_PATH);
		}

		// Set the base path
		$templateInstance->setTemplateBasePath($templateBasePath);

		// Set template extensions
		$templateInstance->setRawTemplateExtension(FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('raw_template_extension'));
		$templateInstance->setCodeTemplateExtension(FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('code_template_extension'));

		// Absolute output path for compiled templates
		$templateInstance->setCompileOutputPath(sprintf('%s%s/',
			$templateBasePath,
			FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('compile_output_path')
		));

		// Return the prepared instance
		return $templateInstance;
	}

	/**
	 * Getter for current main node
	 *
	 * @return	$currMainNode	Current main node
	 */
	public final function getCurrMainNode () {
		return $this->currMainNode;
	}

	/**
	 * Getter for main node array
	 *
	 * @return	$mainNodes	Array with valid main node names
	 */
	public final function getMainNodes () {
		return $this->mainNodes;
	}

	/**
	 * Getter for sub node array
	 *
	 * @return	$subNodes	Array with valid sub node names
	 */
	public final function getSubNodes () {
		return $this->subNodes;
	}

	/**
	 * Setter for image instance
	 *
	 * @param	$imageInstance	An instance of an image
	 * @return	void
	 */
	public final function setImageInstance (BaseImage $imageInstance) {
		$this->imageInstance = $imageInstance;
	}

	/**
	 * Getter for image instance
	 *
	 * @return	$imageInstance	An instance of an image
	 */
	public final function getImageInstance () {
		return $this->imageInstance;
	}

	/**
	 * Handles the start element of an XML resource
	 *
	 * @param	$resource		XML parser resource (currently ignored)
	 * @param	$element		The element we shall handle
	 * @param	$attributes		All attributes
	 * @return	void
	 * @throws	InvalidXmlNodeException		If an unknown/invalid XML node name was found
	 */
	public function startElement ($resource, string $element, array $attributes) {
		// Initial method name which will never be called...
		$methodName = 'initImage';

		// Make the element name lower-case
		$element = strtolower($element);

		// Is the element a main node?
		//* DEBUG: */ echo "START: &gt;".$element."&lt;<br />\n";
		if (in_array($element, $this->mainNodes)) {
			// Okay, main node found!
			$methodName = 'setImage' . StringUtils::convertToClassName($element);
		} elseif (in_array($element, $this->subNodes)) {
			// Sub node found
			$methodName = 'setImageProperty' . StringUtils::convertToClassName($element);
		} elseif ($element != 'image') {
			// Invalid node name found
			throw new InvalidXmlNodeException([$this, $element, $attributes], Parseable::EXCEPTION_XML_NODE_UNKNOWN);
		}

		// Call method
		//* DEBUG: */ echo "call: ".$methodName."<br />\n";
		call_user_func_array(array($this, $methodName), $attributes);
	}

	/**
	 * Ends the main or sub node by sending out the gathered data
	 *
	 * @param	$resource	An XML resource pointer (currently ignored)
	 * @param	$nodeName	Name of the node we want to finish
	 * @return	void
	 * @throws	XmlNodeMismatchException	If current main node mismatches the closing one
	 */
	public function finishElement ($resource, string $nodeName) {
		// Does this match with current main node?
		//* DEBUG: */ echo "END: &gt;".$nodeName."&lt;<br />\n";
		if (($nodeName != $this->getCurrMainNode()) && (in_array($nodeName, $this->getMainNodes()))) {
			// Did not match!
			throw new XmlNodeMismatchException (array($this, $nodeName, $this->getCurrMainNode()), Parseable::EXCEPTION_XML_NODE_MISMATCH);
		} elseif (in_array($nodeName, $this->getSubNodes())) {
			// Silently ignore sub nodes
			return;
		}

		// Construct method name
		$methodName = sprintf('finish%s', StringUtils::convertToClassName($nodeName));

		// Call the corresponding method
		call_user_func_array(array($this->getImageInstance(), $methodName), []);
	}

	/**
	 * Currently not used
	 *
	 * @param	$resource		XML parser resource (currently ignored)
	 * @param	$characters		Characters to handle
	 * @return	void
	 * @todo	Find something usefull with this!
	 */
	public function characterHandler ($resource, string $characters) {
		// Trim all spaces away
		$characters = trim($characters);

		// Is this string empty?
		if (empty($characters)) {
			// Then skip it silently
			return;
		}

		// Unfinished work!
		DebugMiddleware::getSelfInstance()->partialStub('Handling extra characters is not yet supported!');
	}

	/**
	 * Intializes the image
	 *
	 * @return	void
	 * @todo	Add cache creation here
	 */
	private function initImage () {
		// Unfinished work!
	}

	/**
	 * Set the image type
	 *
	 * @param	$imageType	Code fragment or direct value holding the image type
	 * @return	void
	 */
	private function setImageType (string $imageType) {
		// Set group to general
		$this->setVariableGroup('general');

		// Try to compile it first to get the value from variable stack
		$imageType = $this->compileRawCode($imageType);

		// Now make a class name of it
		$className = StringUtils::convertToClassName($imageType.'_image');

		// And try to initiate it
		$this->setImageInstance(ObjectFactory::createObjectByName($className, [$this]));

		// Set current main node to type
		$this->currMainNode = 'type';
	}

	/**
	 * "Setter" for resolution, we first need to collect the resolution from the
	 * sub-nodes. So first, this method will prepare an array for it
	 *
	 * @return	void
	 */
	private function setImageResolution () {
		// Call the image class
		$this->getImageInstance()->initResolution();

		// Current main node is resolution
		$this->currMainNode = 'resolution';
	}

	/**
	 * "Setter" for base information. For more details see above method!
	 *
	 * @return	void
	 * @see		ImageTemplateEngine::setImageResolution
	 */
	private function setImageBase () {
		// Call the image class
		$this->getImageInstance()->initBase();

		// Current main node is resolution
		$this->currMainNode = 'base';
	}

	/**
	 * "Setter" for background-color. For more details see above method!
	 *
	 * @return	void
	 * @see		ImageTemplateEngine::setImageResolution
	 */
	private function setImageBackgroundColor () {
		// Call the image class
		$this->getImageInstance()->initBackgroundColor();

		// Current main node is background-color
		$this->currMainNode = 'background-color';
	}

	/**
	 * "Setter" for foreground-color. For more details see above method!
	 *
	 * @return	void
	 * @see		ImageTemplateEngine::setImageResolution
	 */
	private function setImageForegroundColor () {
		// Call the image class
		$this->getImageInstance()->initForegroundColor();

		// Current main node is foreground-color
		$this->currMainNode = 'foreground-color';
	}

	/**
	 * "Setter" for image-string. For more details see above method!
	 *
	 * @param	$groupable	Whether this image string is groupable
	 * @return	void
	 * @see		ImageTemplateEngine::setImageResolution
	 */
	private function setImageImageString (string $groupable = 'single') {
		// Call the image class
		$this->getImageInstance()->initImageString($groupable);

		// Current main node is foreground-color
		$this->currMainNode = 'image-string';
	}

	/**
	 * Setter for image name
	 *
	 * @param	$imageName	Name of the image
	 * @return	void
	 */
	private function setImagePropertyName (string $imageName) {
		// Call the image class
		$this->getImageInstance()->setImageName($imageName);
	}

	/**
	 * Setter for image width
	 *
	 * @param	$width	Width of the image or variable
	 * @return	void
	 */
	private function setImagePropertyWidth (int $width) {
		// Call the image class
		$this->getImageInstance()->setWidth($width);
	}

	/**
	 * Setter for image height
	 *
	 * @param	$height	Height of the image or variable
	 * @return	void
	 */
	private function setImagePropertyHeight (int $height) {
		// Call the image class
		$this->getImageInstance()->setHeight($height);
	}

	/**
	 * Setter for image red color
	 *
	 * @param	$red	Red color value
	 * @return	void
	 */
	private function setImagePropertyRed ($red) {
		// Call the image class
		$this->getImageInstance()->setRed($red);
	}

	/**
	 * Setter for image green color
	 *
	 * @param	$green	Green color value
	 * @return	void
	 */
	private function setImagePropertyGreen ($green) {
		// Call the image class
		$this->getImageInstance()->setGreen($green);
	}

	/**
	 * Setter for image blue color
	 *
	 * @param	$blue	Blue color value
	 * @return	void
	 */
	private function setImagePropertyBlue ($blue) {
		// Call the image class
		$this->getImageInstance()->setBlue($blue);
	}

	/**
	 * Setter for string name (identifier)
	 *
	 * @param	$stringName		String name (identifier)
	 * @return	void
	 */
	private function setImagePropertyStringName (string $stringName) {
		// Call the image class
		$this->getImageInstance()->setStringName($stringName);
	}

	/**
	 * Setter for font size
	 *
	 * @param	$fontSize	Size of the font
	 * @return	void
	 */
	private function setImagePropertyFontSize (int $fontSize) {
		// Call the image class
		$this->getImageInstance()->setFontSize($fontSize);
	}

	/**
	 * Setter for image string
	 *
	 * @param	$imageString	Image string to set
	 * @return	void
	 */
	private function setImagePropertyText (string $imageString) {
		// Call the image class
		$this->getImageInstance()->setString($imageString);
	}

	/**
	 * Setter for X coordinate
	 *
	 * @param	$x	X coordinate
	 * @return	void
	 */
	private function setImagePropertyX (int $x) {
		// Call the image class
		$this->getImageInstance()->setX($x);
	}

	/**
	 * Setter for Y coordinate
	 *
	 * @param	$y	Y coordinate
	 * @return	void
	 */
	private function setImagePropertyY (int $y) {
		// Call the image class
		$this->getImageInstance()->setY($y);
	}

	/**
	 * Getter for image cache file instance
	 *
	 * @return	$fileInstance	An instance of a SplFileInfo class
	 */
	public function getImageCacheFile () {
		// Get the instance ready
		$fileInstance = new SplFileInfo(sprintf('%s%s%s/%s.%s',
			FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('root_base_path'),
			$this->getGenericBasePath(),
			'images/_cache',
			md5(
				$this->getImageInstance()->getImageName() . ':' . $this->__toString() . ':' . $this->getImageInstance()->__toString()
			),
			$this->getImageInstance()->getImageType()
		));

		// Return it
		return $fileInstance;
	}

	/**
	 * Outputs the image to the world
	 *
	 * @param	$responseInstance	An instance of a Responseable class
	 * @return	void
	 * @todo	Nothing to really "transfer" here?
	 */
	public function transferToResponse (Responseable $responseInstance) {
		// Set the image instance
		$responseInstance->setImageInstance($this->getImageInstance());
	}

	/**
	 * Load a specified image template into the engine
	 *
	 * @param	$template	The image template we shall load which is
	 *						located in 'image' by default
	 * @return	void
	 */
	public function loadImageTemplate (string $template) {
		// Set template type
		$this->setTemplateType(FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('image_template_type'));

		// Load the special template
		$this->loadTemplate($template);
	}

}
